# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import os
import platform
import subprocess
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.logger import get_logger
from scargo.path_utils import find_program_path
from scargo.target_helpers.atsam_helper import (
    AtsamScrips,
    generate_gdb_script,
    generate_openocd_script,
)

if platform.system() == "Windows":
    from subprocess import DETACHED_PROCESS  # type: ignore[attr-defined]
else:
    from subprocess import PIPE

logger = get_logger()


def scargo_flash(
    app: bool,
    fs: bool,
    flash_profile: str,
    port: Optional[str] = None,
    erase_memory: bool = True,
) -> None:
    config = prepare_config()
    target = config.project.target

    if not erase_memory and target.family != "stm32":
        logger.error("--no-erase option is only supported for stm32 projects.")
        sys.exit(1)
    if target.family == "esp32":
        flash_esp32(config, app=app, fs=fs, flash_profile=flash_profile, port=port)
    elif target.family == "stm32":
        flash_stm32(config, flash_profile, erase_memory, port=port)
    elif target.family == "atsam":
        flash_atsam(config, flash_profile)
    else:
        logger.error("Flash command not supported for this target yet.")


def flash_esp32(
    config: Config,
    app: bool,
    fs: bool,
    flash_profile: str = "Debug",
    port: Optional[str] = None,
) -> None:
    project_path = config.project_root
    out_dir = project_path / "build" / flash_profile
    chip = config.get_esp32_config().chip
    command = []
    try:
        if app:
            app_name = config.project.name
            app_path = out_dir / f"{app_name}.bin"
            command = ["parttool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(
                [
                    "write_partition",
                    "--partition-name=ota_0",
                    f"--input={app_path}",
                ]
            )
            subprocess.check_call(command)
        elif fs:
            fs_path = config.project_root / "build" / "spiffs.bin"
            command = ["parttool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(
                [
                    "write_partition",
                    "--partition-name=spiffs",
                    f"--input={fs_path}",
                ]
            )
            subprocess.check_call(command)
        else:
            command = ["esptool.py"]
            if port:
                command.append(f"--port={port}")
            command.extend(["--chip", chip, "write_flash", "@flash_args"])
            subprocess.check_call(command, cwd=out_dir)
    except subprocess.CalledProcessError:
        logger.error("%s fail", command)


def flash_stm32(
    config: Config,
    flash_profile: str = "Debug",
    erase_memory: bool = True,
    port: Optional[str] = None,
) -> None:
    project_path = config.project_root
    bin_path = (
        project_path
        / "build"
        / flash_profile
        / "bin"
        / f"{config.project.name.lower()}.bin"
    )

    flash_start = hex(config.get_stm32_config().flash_start)

    if not bin_path.exists():
        logger.error("%s does not exist", bin_path)
        logger.info("Did you run scargo build --profile %s", flash_profile)
    elif not flash_start:
        logger.error("Flash start address not found in lock file")
        logger.info("Define flash-start in scargo.toml under stm32 section")
    else:
        if erase_memory:
            command = ["st-flash"]
            if port:
                command.append(f"--serial={port}")
            command.append("erase")
            subprocess.check_call(command)

        command = ["st-flash", "--reset"]
        if port:
            command.append(f"--serial={port}")
        command.extend(["write", str(bin_path), flash_start])
        subprocess.check_call(command)


def flash_atsam(
    config: Config,
    flash_profile: str = "Debug",
) -> None:
    openocd_path = find_program_path("openocd")
    gdb_multiarch_path = find_program_path("gdb-multiarch")

    if not openocd_path:
        logger.error("openocd not found")
        logger.info("Please install openocd")
        sys.exit(1)

    if not gdb_multiarch_path:
        logger.error("gdb-multiarch not found")
        logger.info("Please install gdb-multiarch")
        sys.exit(1)

    project_path = config.project_root
    bin_name = f"{config.project.name.lower()}.bin"
    build_path = Path(project_path, "build", flash_profile, "bin")
    bin_path = build_path / bin_name
    exec_path = build_path / config.project.name.lower()

    if not bin_path.exists():
        logger.error("%s does not exist", bin_path)
        logger.info("Did you run scargo build --profile %s", flash_profile)
        sys.exit(1)

    temp_script_dir = TemporaryDirectory()
    temp_script_dir_path = Path(temp_script_dir.name)
    generate_openocd_script(temp_script_dir_path, config)
    generate_gdb_script(temp_script_dir_path, config, bin_path)

    openocd_process = None
    try:
        if platform.system() == "Windows":
            openocd_process = subprocess.Popen(
                [
                    openocd_path,
                    "-f",
                    str(temp_script_dir_path / AtsamScrips.openocd_cfg),
                ],
                creationflags=DETACHED_PROCESS,
            )
        else:
            openocd_process = subprocess.Popen(
                [
                    "sudo",
                    openocd_path,
                    "-f",
                    str(temp_script_dir_path / AtsamScrips.openocd_cfg),
                ],
                stdin=PIPE,
                stdout=PIPE,
                stderr=PIPE,
            )

        gdb_command = [
            str(gdb_multiarch_path),
            f"{exec_path}",
            f"--command={temp_script_dir_path / AtsamScrips.gdb_flash}",
            "--batch",
        ]
        subprocess.check_call(gdb_command)
    except Exception as e:
        print(e)
    finally:
        # temp_script_dir.cleanup()
        if openocd_process is not None:
            if platform.system() == "Windows":
                openocd_process.terminate()
            else:
                os.system("sudo pkill -9 -P " + str(openocd_process.pid))



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from pathlib import Path
from typing import Any, Dict

import yaml

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.logger import get_logger

logger = get_logger()


class _CicdTemplate:
    """
    This class is a container for cicd yaml files creation with multilayer approach
    """

    def __init__(self, config: Config):
        self._config = config

    @classmethod
    def merge_dictionaries(
        cls, dict1: Dict[str, Any], dict2: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Marge two dictionaries recursively

        Args:
            dict1 (dict): base_dict
            dict2 (dict): custom_dict

        Returns:
            dict: merged dictionary
        """
        for key in dict2:
            if (
                key in dict1
                and isinstance(dict1[key], dict)
                and isinstance(dict2[key], dict)
            ):
                cls.merge_dictionaries(dict1[key], dict2[key])
            else:
                dict1[key] = dict2[key]
        return dict1

    def _get_cicd_content(self, path: str, file_name: str) -> Dict[str, Any]:
        ci_folder = Path(path)
        file_path = ci_folder / file_name

        # Check if the ci folder exists. If not, create it.
        if not ci_folder.exists():
            ci_folder.mkdir(parents=True)

        # Check if file exists. If not, create it.
        if not file_path.exists():
            file_path.touch()

        # Read the file
        with file_path.open("r", encoding="utf-8") as file:
            data = yaml.safe_load(file)

        if not isinstance(data, dict):
            return {}
        return data

    def generate_cicd_env(self) -> None:
        """Generate dirs and files"""
        create_file_from_template(
            ".gitlab-ci-custom.yml.j2",
            ".devcontainer/.gitlab-ci-custom.yml",
            template_params={},
            config=self._config,
            overwrite=False,
        )

        create_file_from_template(
            ".gitlab-ci.yml.j2",
            ".gitlab-ci.yml",
            template_params={
                "project": self._config.project,
            },
            config=self._config,
        )
        custom_cicd = self._get_cicd_content(".devcontainer", ".gitlab-ci-custom.yml")

        # Check if custom_cicd is empty and if so, do not merge
        if custom_cicd is None:
            logger.info("Custom file is empty or currupted. Skipping cicd file merge.")
            return

        # Merge custom_cicd with base_cicd
        base_cicd = self._get_cicd_content(".", ".gitlab-ci.yml")
        merged_dict = self.merge_dictionaries(base_cicd, custom_cicd)
        with open(".gitlab-ci.yml", "w", encoding="utf-8") as f:
            yaml.dump(merged_dict, f, sort_keys=False)


def generate_cicd(config: Config) -> None:
    """Generate cicd file with custom user layer

    Args:
        config (Config): target configuration
    """
    cicd_template = _CicdTemplate(config)
    cicd_template.generate_cicd_env()



import xml.etree.ElementTree as ET
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from scargo.config import Config
from scargo.file_generators.base_gen import write_template

script_dir = Path(__file__).parent
atmel_arxml_path = script_dir / "atmel.xml"

ADDITIONAL_CPU_DATA = {"cortex-m23": ["atsaml10e16a"]}


@dataclass
class AtsamScrips:
    openocd_cfg = "config/openocd_script.cfg"
    gdb_flash = "config/gdb-flash.script"


def get_atsam_cpu(chip_label: str) -> Optional[str]:
    cpu = None
    chip_label = chip_label.lower()

    if not chip_label.startswith("atsam"):
        return None

    # Check in dict
    for cpu, chip_list in ADDITIONAL_CPU_DATA.items():
        if chip_label in chip_list:
            return cpu

    # Check in xml
    tree = ET.parse(atmel_arxml_path)
    root = tree.getroot()
    for element in root.iter():
        name = element.attrib.get("name", "")
        if name.startswith("cortex"):
            cpu = name
        if chip_label in name:
            return cpu
    return None


def get_openocd_script_name(chip_label: str) -> Optional[str]:
    if chip_label.startswith("atsamd"):
        return "at91samdXX.cfg"
    elif chip_label.startswith("atsaml1"):
        return "atsaml1x.cfg"  # Not sure about this
    return None


def get_openocd_flash_driver_name(chip_label: str) -> Optional[str]:
    # https://openocd.org/doc-release/html/Flash-Commands.html
    # This should probably be done differently
    if chip_label.startswith("atsamd"):
        return "at91samd"
    elif chip_label.startswith("atsaml1"):
        return "at91samd"
    return None


def generate_openocd_script(outdir: Path, config: Config) -> None:
    openocd_script_name = get_openocd_script_name(
        config.get_atsam_config().chip.lower()
    )
    flash_driver = get_openocd_flash_driver_name(config.get_atsam_config().chip.lower())
    if openocd_script_name and openocd_script_name:
        write_template(
            outdir / AtsamScrips.openocd_cfg,
            "atsam/openocd_script.cfg.j2",
            {"flash_driver": flash_driver, "script_name": openocd_script_name},
        )


def generate_gdb_script(outdir: Path, config: Config, bin_path: Path) -> None:
    flash_driver = get_openocd_flash_driver_name(config.get_atsam_config().chip.lower())
    if flash_driver:
        write_template(
            outdir / AtsamScrips.gdb_flash,
            "atsam/gdb-flash.script.j2",
            {"flash_driver": flash_driver, "bin_path": bin_path},
        )
